#include "model/CBitStorage.h"

#include "ui/CLog.h"
#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"

#include <windows.h>

// L2CacheD inlined in CUser c.tor
CBitStorage::CBitStorage()
{
    guard(L"CBitStorage()", L"CBitStorage() (exit)", TEXT(__FILE__), __LINE__);
    ::memset(m_data, 0, sizeof(m_data));
    unguard();  // base class
}

// L2CacheD inlined
CBitStorage::~CBitStorage()
{
}

int CBitStorage::IdToIndex(int bitIndex) const
{
    guard(L"inline int IdToIndex(int bitIndex) const", L"inline int IdToIndex(int bitIndex) const (exit)", TEXT(__FILE__), __LINE__);

    int bitPosition = 1 << (bitIndex & 0x1F);

    unguard();

    return bitPosition;
}

// L2CacheD 0x00460500
void CBitStorage::SetFlag(int bitIndex, bool sw)
{
    guard(L"SetFlag(int nId, bool bSw)", L"SetFlag(int nId, bool bSw) (exit)", TEXT(__FILE__), __LINE__);
    if (bitIndex >= 256)
    {
        g_winlog.Add(LOG_ERROR, L"[%s][%d] Bit flag Index overflow", TEXT(__FILE__), __LINE__);
        unguard();
        return;
    }

    if (bitIndex < 0)
    {
        g_winlog.Add(LOG_ERROR, L"[%s][%d] Bit flag index underflow", TEXT(__FILE__), __LINE__);
        unguard();
        return;
    }

    const int indexInInteger = IdToIndex(bitIndex);
    int chunkForBit = m_data[bitIndex >> 5];
    int newValue = 0;
    if (sw)
    {
        newValue = indexInInteger | chunkForBit;
    }
    else
    {
        newValue = ~indexInInteger & chunkForBit;
    }

    m_data[bitIndex >> 5] = newValue;

    unguard();
}

// L2CacheD inlined
void CBitStorage::GetFlagAll(uint8_t* flags)
{
    guard(L"GetFlagAll(char * sFlag)", L"GetFlagAll(char * sFlag) (exit)", TEXT(__FILE__), __LINE__);
    ::memcpy(flags, m_data, sizeof(m_data));
    unguard();
}