#include "model/CItemData.h"

#include "db/DBConn.h"
#include "ui/CLog.h"
#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"

CItemData g_itemData;  // L2CacheD 0x0268493C

// L2CacheD 0x00482E70
CItemData::CItemData()
    : m_lock()
    , m_types()
{
    guard(L"CItemData::CItemData()", L"CItemData::CItemData() (exit)", TEXT(__FILE__), __LINE__);
    ::InitializeCriticalSection(&m_lock);
    unguard();
}

// L2CacheD 0x0049D9B0
CItemData::~CItemData()
{
    guard(L"CItemData::~CItemData()", L"CItemData::~CItemData() (exit)", TEXT(__FILE__), __LINE__);
    ::DeleteCriticalSection(&m_lock);
    unguard();
}

// L2CacheD 0x00482BE0
ItemConsumeType CItemData::GetConsumeType(int itemType)
{
    guard(L"int CItemData::GetConsumeType(int nItemType)", L"int CItemData::GetConsumeType(int nItemType) (exit)", TEXT(__FILE__), __LINE__);

    ::EnterCriticalSection(&m_lock);

    ItemConsumeTypes::iterator it = m_types.find(itemType);
    if (it != m_types.end())
    {
        ItemConsumeType result = it->second;
        ::LeaveCriticalSection(&m_lock);

        unguard();
        return result;
    }

    wchar_t consumeTypeStr[50];
    ::memset(consumeTypeStr, 0, sizeof(consumeTypeStr));

    DBConn sql;
    sql.Bind(consumeTypeStr, sizeof(consumeTypeStr));
    if (!sql.Execute(L"EXEC lin_GetItemData %d", itemType))
    {
        ::LeaveCriticalSection(&m_lock);
        unguard();
        return ConsumeType_Invalid;
    }

    if (!sql.Fetch())
    {
        ::LeaveCriticalSection(&m_lock);
        unguard();
        return ConsumeType_Invalid;
    }

    ItemConsumeType result = ConsumeType_Invalid;
    if (::wcscmp(consumeTypeStr, L"consume_type_normal") == 0)
    {
        result = ConsumeType_Normal;
    }
    if (::wcscmp(consumeTypeStr, L"consume_type_asset") == 0)
    {
        result = ConsumeType_Asset;
    }
    if (::wcscmp(consumeTypeStr, L"consume_type_charge") == 0)
    {
        result = ConsumeType_Charge;
    }
    if (::wcscmp(consumeTypeStr, L"consume_type_stackable") == 0)
    {
        result = ConsumeType_Stackable;
    }

    ::LeaveCriticalSection(&m_lock);

    unguard();
    return result;
}
