#include "model/CPledgeDb.h"

#include "db/DBConn.h"
#include "config/Config.h"
#include "model/CPledgeWar.h"
#include "model/CPledgeWarDb.h"
#include "ui/CLog.h"
#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"
#include "utils/Utils.h"

CPledgeDb g_pledgeDb;  // L2CacheD 0x019E6368

// L2CacheD 0x004306D0
CPledgeDb::CPledgeDb()
    : m_lock()
    , m_pledgesById()
    , m_pledgesByName()
{
    guard(L"CPledgeDb::CPledgeDb()", L"CPledgeDb::CPledgeDb() (exit)", TEXT(__FILE__), __LINE__);
    unguard();
}

// L2CacheD 0x00430760
CPledgeDb::~CPledgeDb()
{
    guard(L"CPledgeDb::~CPledgeDb()", L"CPledgeDb::~CPledgeDb() (exit)", TEXT(__FILE__), __LINE__);
    unguard();
}

// L2CacheD 0x004307F0
void CPledgeDb::Init()
{
    guard(L"CPledgeDb::Init()", L"CPledgeDb::Init() (exit)", TEXT(__FILE__), __LINE__);

    uint32_t id = 0;
    uint32_t beginTime = 0;
    uint32_t challenger = 0;
    uint32_t challengee = 0;

    DBConn sql;
    sql.Bind(&id);
    sql.Bind(&beginTime);
    sql.Bind(&challenger);
    sql.Bind(&challengee);
    bool execResult = sql.Execute(L"EXEC lin_LoadAllWarData %d", 0);
    if (execResult)
    {
        while (sql.Fetch())
        {
            g_pledgeWarsDb.GetPledgeWar(id);
            GetPledge(challenger);
            GetPledge(challengee);
        }
    }

    unguard();
}

// L2CacheD 0x0042E630
CPledgeSP CPledgeDb::LoadPledge(const wchar_t* name)
{
    guard(L"CPledgeSP CPledgeDb::LoadPledge(wchar_t* sPledgeName)", L"CPledgeSP CPledgeDb::LoadPledge(wchar_t* sPledgeName) (exit)", TEXT(__FILE__), __LINE__);

    uint32_t pledgeId = 0;
    wchar_t nameOut[24];
    ::memset(nameOut, 0, sizeof(nameOut));  // FIXED: added
    uint32_t rulerId = 0;
    wchar_t rulerName[24];
    ::memset(rulerName, 0, sizeof(rulerName));  // FIXED: added
    uint32_t allianceId = 0;
    int challengeTime = 0;
    uint32_t nowWarId = 0;
    int nameValue = 0;
    int oustTime = 0;
    int skillLevel = 0;
    int privateFlag = 0;
    int status = 0;
    int rank = 0;
    uint32_t castleId = 0;
    uint32_t agitId = 0;
    int rootNameValue = 0;
    uint32_t crestId = 0;
    int isGuilty = 0;
    int dismissReservedTime = 0;
    int allianceOustedTime = 0;
    int allianceWithdrawTime = 0;
    int allianceDismissTime = 0;

    SQLINTEGER length = -1;

    DBConn sql;
    sql.Bind(&pledgeId);
    sql.Bind(nameOut, sizeof(nameOut));
    sql.Bind(&rulerId);
    sql.Bind(rulerName, sizeof(rulerName));
    sql.BindNullable(&allianceId, &length);
    sql.BindNullable(&challengeTime, &length);
    sql.BindNullable(&nowWarId, &length);
    sql.BindNullable(&nameValue, &length);
    sql.BindNullable(&oustTime, &length);
    sql.BindNullable(&skillLevel, &length);
    sql.BindNullable(&privateFlag, &length);
    sql.BindNullable(&status, &length);
    sql.BindNullable(&rank, &length);
    sql.BindNullable(&castleId, &length);
    sql.BindNullable(&agitId, &length);
    sql.BindNullable(&rootNameValue, &length);
    sql.BindNullable(&crestId, &length);
    sql.BindNullable(&isGuilty, &length);
    sql.BindNullable(&dismissReservedTime, &length);
    sql.BindNullable(&allianceOustedTime, &length);
    sql.BindNullable(&allianceWithdrawTime, &length);
    sql.BindNullable(&allianceDismissTime, &length);
    bool execResult = sql.Execute(L"EXEC lin_LoadPledgeByName N'%s'", name);

    if (!execResult)
    {
        unguard();
        return CPledgeSP();
    }

    if (!sql.Fetch())
    {
        unguard();
        return CPledgeSP();
    }

    CPledge* pledge = new CPledge(pledgeId, name, rulerId, rulerName, allianceId, challengeTime, nowWarId, nameValue, oustTime, skillLevel, privateFlag, status, rank, castleId, agitId, rootNameValue, 0, crestId, isGuilty, dismissReservedTime, allianceOustedTime, allianceWithdrawTime, allianceDismissTime);

    LoadPledgeMember(pledge);
    // InsertIntoMap(-1, name, pledge); FIXED
    InsertIntoMap(pledgeId, name, pledge);
    pledge->Release(__FILE__, __LINE__, ORT_CREATE_DELETE, 0);

    CPledgeSP result = m_pledgesById.FindObject(pledgeId);

    unguard();
    return result;
}

// L2CacheD 0x0042EAA0
CPledgeSP CPledgeDb::LoadPledge(uint32_t pledgeId)
{
    guard(L"CPledgeSP CPledgeDb::LoadPledge(int nPledgeId)", L"CPledgeSP CPledgeDb::LoadPledge(int nPledgeId) (exit)", TEXT(__FILE__), __LINE__);

    uint32_t pledgeIdOut = 0;
    wchar_t name[24];
    ::memset(name, 0, sizeof(name));  // FIXED: added
    uint32_t rulerId = 0;
    wchar_t rulerName[24];
    ::memset(rulerName, 0, sizeof(rulerName));  // FIXED: added
    uint32_t allianceId = 0;
    int challengeTime = 0;
    uint32_t nowWarId = 0;
    int nameValue = 0;
    int oustTime = 0;
    int skillLevel = 0;
    int privateFlag = 0;
    int status = 0;
    int rank = 0;
    uint32_t castleId = 0;
    uint32_t agitId = 0;
    int rootNameValue = 0;
    uint32_t crestId = 0;
    int isGuilty = 0;
    int dismissReservedTime = 0;
    int allianceOustedTime = 0;
    int allianceWithdrawTime = 0;
    int allianceDismissTime = 0;

    SQLINTEGER length = -1;

    DBConn sql;
    sql.Bind(&pledgeIdOut);
    sql.Bind(name, sizeof(name));
    sql.Bind(&rulerId);
    sql.Bind(rulerName, sizeof(rulerName));
    sql.BindNullable(&allianceId, &length);
    sql.BindNullable(&challengeTime, &length);
    sql.BindNullable(&nowWarId, &length);
    sql.BindNullable(&nameValue, &length);
    sql.BindNullable(&oustTime, &length);
    sql.BindNullable(&skillLevel, &length);
    sql.BindNullable(&privateFlag, &length);
    sql.BindNullable(&status, &length);
    sql.BindNullable(&rank, &length);
    sql.BindNullable(&castleId, &length);
    sql.BindNullable(&agitId, &length);
    sql.BindNullable(&rootNameValue, &length);
    sql.BindNullable(&crestId, &length);
    sql.BindNullable(&isGuilty, &length);
    sql.BindNullable(&dismissReservedTime, &length);
    sql.BindNullable(&allianceOustedTime, &length);
    sql.BindNullable(&allianceWithdrawTime, &length);
    sql.BindNullable(&allianceDismissTime, &length);
    bool execResult = sql.Execute(L" EXEC lin_LoadPledgeById %d", pledgeId);

    if (!execResult)
    {
        unguard();
        return CPledgeSP();
    }

    if (!sql.Fetch())
    {
        unguard();
        return CPledgeSP();
    }

    CPledge* pledge = new CPledge(pledgeId, name, rulerId, rulerName, allianceId, challengeTime, nowWarId, nameValue, oustTime, skillLevel, privateFlag, status, rank, castleId, agitId, rootNameValue, 0, crestId, isGuilty, dismissReservedTime, allianceOustedTime, allianceWithdrawTime, allianceDismissTime);

    LoadPledgeMember(pledge);
    InsertIntoMap(pledgeId, name, pledge);

    pledge->Release(__FILE__, __LINE__, ORT_CREATE_DELETE, 0);

    CPledgeSP result = m_pledgesById.FindObject(pledgeId);

    unguard();
    return result;
}

// L2CacheD 0x0042EF00
CPledgeSP CPledgeDb::GetPledge(const wchar_t* pledgeName)
{
    guard(L"CPledgeSP CPledgeDb::GetPledge(wchar_t* sPledgeName)", L"CPledgeSP CPledgeDb::GetPledge(wchar_t* sPledgeName) (exit)", TEXT(__FILE__), __LINE__);

    wchar_t pledgeNameLowerCase[26];
    ::memset(pledgeNameLowerCase, 0, sizeof(pledgeNameLowerCase));  // FIXED: added
    Utils::CopyStrInLowerCase(pledgeName, pledgeNameLowerCase);

    m_lock.WriteLock();

    CPledgeSP result = m_pledgesByName.FindObject(std::wstring(pledgeNameLowerCase));
    if (result == NULL)
    {
        result = LoadPledge(pledgeName);
    }

    m_lock.WriteUnlock();

    unguard();
    return result;
}

// L2CacheD 0x0042F080
CPledgeSP CPledgeDb::GetPledge(uint32_t pledgeId)
{
    guard(L"CPledgeSP CPledgeDb::GetPledge(int nPledgeId)", L"CPledgeSP CPledgeDb::GetPledge(int nPledgeId) (exit)", TEXT(__FILE__), __LINE__);

    m_lock.WriteLock();

    CPledgeSP result = m_pledgesById.FindObject(pledgeId);
    if (result == NULL)
    {
        result = LoadPledge(pledgeId);
    }

    m_lock.WriteUnlock();

    unguard();
    return result;
}

// L2CacheD 0x0042FB90
int CPledgeDb::CreatePledge(const wchar_t* pledgeName, uint32_t rulerId)
{
    guard(L"int CPledgeDb::CreatePledge(wchar_t *sPledgeName, int nPledgeRulerId)", L"int CPledgeDb::CreatePledge(wchar_t *sPledgeName, int nPledgeRulerId) (exit)", TEXT(__FILE__), __LINE__);

    CPledgeSP existed = GetPledge(pledgeName);
    if (existed != NULL)
    {
        unguard();
        return -1;
    }

    uint32_t pledgeId = 0;
    wchar_t rullerName[24];
    DBConn sql;
    sql.Bind(&pledgeId);
    sql.Bind(rullerName, sizeof(rullerName));

    bool execResult = sql.Execute(L"EXEC lin_CreatePledge N'%s', %d", pledgeName, pledgeName, rulerId);

    if (!execResult)
    {
        unguard();
        return -1;
    }

    if (!sql.Fetch())
    {
        unguard();
        return -1;
    }

    if (pledgeId == 0)
    {
        unguard();
        return -1;
    }

    m_lock.WriteLock();

    CPledge* pledge = new CPledge(pledgeId, pledgeName, rulerId, rullerName);
    pledge->AddMember(rulerId);
    InsertIntoMap(pledgeId, pledgeName, pledge);

    m_lock.WriteUnlock();

    pledge->Release(__FILE__, __LINE__, ORT_CREATE_DELETE, 0);

    return pledgeId;
}

// L2CacheD 0x0042FDF0
void CPledgeDb::DeletePledge(uint32_t pledgeId)
{
    DBConn sql;
    if (!sql.Execute(L"EXEC lin_DeletePledge %d", pledgeId))
    {
        return;
    }

    CPledgeSP pledge = GetPledge(pledgeId);
    if (pledge != NULL)
    {
        g_winlog.Add(LOG_ERROR, L"cannot find pledge. dbid[%d]", pledgeId);
        return;
    }

    DeleteFromMap(pledge.get());

    //    begin = pledge->set_membersSet.base._Myhead->_Left;
    //    it = begin;
    //    if (it != pledge->set_membersSet.base._Myhead)
    //    {
    //        do
    //        {
    //            CUserDb::GetUser_byId(&g_userDb, &userPtr, it->_Myval, true); TODO: PORT ME
    //            LOBYTE(v12) = 2;
    //            if (userPtr.m_object)
    //            {
    //                user = CSPointer::operator_indirect(&userPtr);
    //                CUser::SetPledge(user, 0);
    //            }
    //            else
    //            {
    //                g_winlog.Add(LOG_ERROR, L"[%S][%d]Cannot find user, id[%d]", "PledgeDb.cpp", 657, it->_Myval);
    //            }
    //            LOBYTE(v12) = 1;
    //            if (userPtr.m_object)
    //            {
    //                ((void(__stdcall*)(const char*, signed int, ORTs, _DWORD))userPtr.m_object->vtable->Release)("c:\\serverservice\\cached\\ThreadSafeLibrary.h", 119, userPtr.m_workType, 0);
    //                userPtr.m_object = 0;
    //            }
    //            set::iterator::increment(&begin);
    //            v6 = CPledgeSP_or_CAllianceSP::operator_indirect(&pledge);
    //            it = begin;
    //        } while (begin != v6->m_d.m_membersSet.base._Myhead);
    //    }
}

// L2CacheD 0x0042E3C0
void CPledgeDb::InsertIntoMap(uint32_t pledgeId, const wchar_t* name, CPledge* pledge)
{
    guard(L"void CPledgeDb::InsertIntoMap(uint32_t nPledgeId, wchar_t *sPledgeName, CPledge *pPledge)", L"void CPledgeDb::InsertIntoMap(uint32_t nPledgeId, wchar_t *sPledgeName, CPledge *pPledge) (exit)", TEXT(__FILE__), __LINE__);

    wchar_t pledgeNameLowerCase[26];
    ::memset(pledgeNameLowerCase, 0, sizeof(pledgeNameLowerCase));  // FIXED: added
    Utils::CopyStrInLowerCase(name, pledgeNameLowerCase);

    m_pledgesById.Insert(pledgeId, pledge, NULL, -1);
    m_pledgesByName.Insert(std::wstring(pledgeNameLowerCase), pledge, NULL, -1);

    unguard();
}

// L2CacheD 0x0042E4C0
void CPledgeDb::DeleteFromMap(CPledge* pledge)
{
    guard(L"void CPledgeDb::DeleteFromMap(CPledge *pPledge)", L"void CPledgeDb::DeleteFromMap(CPledge *pPledge) (exit)", TEXT(__FILE__), __LINE__);

    wchar_t pledgeName[26];
    ::wcscpy(pledgeName, pledge->name());

    wchar_t pledgeNameLowerCase[26];
    ::memset(pledgeNameLowerCase, 0, sizeof(pledgeNameLowerCase));  // FIXED: added
    Utils::CopyStrInLowerCase(pledgeName, pledgeNameLowerCase);

    m_lock.WriteLock();

    m_pledgesById.DeleteObject(pledge->id(), NULL, -1);
    m_pledgesByName.DeleteObject(std::wstring(pledgeNameLowerCase), NULL, -1);

    m_lock.WriteUnlock();

    unguard();
}

// L2CacheD 0x0042C8D0
bool CPledgeDb::SetUserPledgeInfo(uint32_t userId, PledgeMemberFields updateField, uint32_t updateValue)
{
    guard(L"bool CPledgeDb::SetUserPledgeInfo(uint32_t nUserId, uint32_t nUpdateFieldName, uint32_t nUpdateValue)", L"bool CPledgeDb::SetUserPledgeInfo(uint32_t nUserId, uint32_t nUpdateFieldName, uint32_t nUpdateValue) (exit)", TEXT(__FILE__), __LINE__);

    const wchar_t* fieldName = NULL;
    switch (updateField)
    {
        case PledgeMemberField_PledgeDismissTime:
            fieldName = L"pledge_dismiss_time";
            break;
        case PledgeMemberField_PledgeOustedTime:
            fieldName = L"pledge_ousted_time";
            break;
        case PledgeMemberField_PledgeWithdrawTime:
            fieldName = L"pledge_withdraw_time";
            break;
        case PledgeMemberField_SurrenderWarId:
            fieldName = L"surrender_war_id";
            break;
        default:
            unguard();
            return false;
    }

    DBConn sql;
    if (!sql.Execute(L"EXEC lin_UpdatePledgeInfo N'%s', %d, %d", fieldName, updateValue, userId))
    {
        g_winlog.Add(LOG_ERROR, L"Error in Updating Field user_data[%d]", userId);
        unguard();
        return false;
    }

    //    CUserSP user = g_userDb.GetUser(userId, true); TODO: port me
    //    if (user != NULL)
    //    {
    //        switch (updateField)
    //        {
    //            case PledgeMemberField_PledgeDismissTime:
    //                user->setPledgeDismissTime(updateValue);
    //                break;
    //            case PledgeMemberField_PledgeOustedTime:
    //                user->setPledgeOustedTime(updateValue);
    //                break;
    //            case PledgeMemberField_PledgeWithdrawTime:
    //                user->setPledgeWithdrawTime(updateValue);
    //                break;
    //            case PledgeMemberField_SurrenderWarId:
    //                user->setSurrenderWarId(updateValue);
    //                break;
    //            default:
    //                unguard();
    //                return false;
    //        }
    //    }

    unguard();
    return true;
}

// L2CacheD 0x0042F1A0
bool CPledgeDb::SetPledgeInfo(uint32_t pledgeId, PledgeFields updateField, uint32_t updateValue)
{
    guard(L"bool CPledgeDb::SetPledgeInfo(uint32_t nPledgeId, uint32_t nUpdateFieldName, uint32_t nUpdateValue)", L"bool CPledgeDb::SetPledgeInfo(uint32_t nPledgeId, uint32_t nUpdateFieldName, uint32_t nUpdateValue) (exit)", TEXT(__FILE__), __LINE__);

    const wchar_t* updateFieldName = NULL;
    switch (updateField)
    {
        case PledgeField_AllianceId:
            updateFieldName = L"alliance_id";
            break;
        case PledgeField_ChallengeTime:
            updateFieldName = L"challenge_time";
            break;
        case PledgeField_RootNameValue:
            updateFieldName = L"root_name_value";
            break;
        case PledgeField_NowWarId:
            updateFieldName = L"now_war_id";
            break;
        case PledgeField_OustTime:
            updateFieldName = L"oust_time";
            break;
        case PledgeField_SkillLevel:
            updateFieldName = L"skill_level";
            break;
        case PledgeField_CastleId:
            updateFieldName = L"castle_id";
            break;
        case PledgeField_AgitId:
            updateFieldName = L"agit_id";
            break;
        case PledgeField_Rank:
            updateFieldName = L"rank";
            break;
        case PledgeField_NameValue:
            updateFieldName = L"name_value";
            break;
        case PledgeField_Status:
            updateFieldName = L"status";
            break;
        case PledgeField_PrivateFlag:
            updateFieldName = L"private_flag";
            break;
        case PledgeField_CrestId:
            updateFieldName = L"crest_id";
            break;
        case PledgeField_IsGuilty:
            updateFieldName = L"is_guilty";
            break;
        case PledgeField_DismissReservedTime:
            updateFieldName = L"dismiss_reserved_time";
            break;
        case PledgeField_AllianceOustedTime:
            updateFieldName = L"alliance_ousted_time";
            break;
        case PledgeField_AllianceWithdrawTime:
            updateFieldName = L"alliance_withdraw_time";
            break;
        case PledgeField_AllianceDismissTime:
            updateFieldName = L"alliance_dismiss_time";
            break;
        case PledgeField_RulerId:
            updateFieldName = L"ruler_id";
            break;
        default:
            unguard();
            return false;
    }

    DBConn sql;
    if (!sql.Execute(L"EXEC lin_SetPledgeInfo N'%s', %d, %d", updateFieldName, updateValue, pledgeId))
    {
        g_winlog.Add(LOG_ERROR, L"Error in Updating Field Pledge[%d]", pledgeId);
        unguard();
        return false;
    }

    CPledgeSP pledge = GetPledge(pledgeId);
    if (pledge != NULL)
    {
        switch (updateField)
        {
            case PledgeField_AllianceId:
                pledge->setAllianceId(updateValue);
                break;
            case PledgeField_ChallengeTime:
                pledge->setChallengeTime(updateValue);
                break;
            case PledgeField_RootNameValue:
                pledge->setRootNameValue(updateValue);
                break;
            case PledgeField_NowWarId:
                pledge->setNowWarId(updateValue);
                break;
            case PledgeField_OustTime:
                pledge->setOustTime(updateValue);
                break;
            case PledgeField_SkillLevel:
                pledge->setSkillLevel(updateValue);
                break;
            case PledgeField_CastleId:
                pledge->setCastleId(updateValue);
                break;
            case PledgeField_AgitId:
                pledge->setAgitId(updateValue);
                break;
            case PledgeField_Rank:
                pledge->setRank(updateValue);
                break;
            case PledgeField_NameValue:
                pledge->setNameValue(updateValue);
                break;
            case PledgeField_Status:
                pledge->setStatus(updateValue);
                break;
            case PledgeField_PrivateFlag:
                pledge->setPrivateFlag(updateValue);
                break;
            case PledgeField_CrestId:
                pledge->setCrestId(updateValue);
                break;
            case PledgeField_IsGuilty:
                pledge->setIsGuilty(updateValue);
                break;
            case PledgeField_DismissReservedTime:
                pledge->setDismissReservedTime(updateValue);
                break;
            case PledgeField_AllianceOustedTime:
                pledge->setAllianceOustedTime(updateValue);
                break;
            case PledgeField_AllianceWithdrawTime:
                pledge->setAllianceWithdrawTime(updateValue);
                break;
            case PledgeField_AllianceDismissTime:
                pledge->setAllianceDismissTime(updateValue);
                break;
            case PledgeField_RulerId:
                pledge->setRullerId(updateValue);
                break;
        }
    }

    if (updateField != PledgeField_Status)
    {
        unguard();
        return true;
    }

    if (pledge == NULL)
    {
        unguard();
        return true;
    }

    wchar_t pledgeName[26];
    ::memset(pledgeName, 0, sizeof(pledgeName));
    ::wcscpy(pledgeName, pledge->name());

    //    CUserSP user = g_userDb.GetUser(pledge->rullerId(), true); TODO: port me
    //    if (user == NULL)
    //    {
    //        g_winlog.Add(LOG_ERROR, L"Error in Updating Field Pledge[%d]", pledgeId);
    //        unguard();
    //        return false;
    //    }

    //    if (updateValue == 3)
    //    {
    //        charId = CSPointer::operator_indirect(&rullerPtr)->m_d.m_charId_3;
    //        v10 = CSPointer::operator_indirect(&rullerPtr);
    //        accountName = CUser::GetAccountName(v10);
    //        v11 = CSPointer::operator_indirect(&rullerPtr);
    //        charName = CUser::GetCharName(v11);
    //        pledgeId2 = pledgeId;
    //        v12 = CSPointer::operator_indirect(&rullerPtr);
    //        level = CUser::GetLevel(v12);
    //        v13 = CSPointer::operator_indirect(&rullerPtr);
    //        class_1 = CUser::GetClass(v13);
    //        v14 = CSPointer::operator_indirect(&rullerPtr);
    //        gender = CUser::GetGender(v14);
    //        v15 = CSPointer::operator_indirect(&rullerPtr);
    //        race = CUser::GetRace(v15);
    //        pledgeName2 = pledgeName;
    //        v16 = CSPointer::operator_indirect(&rullerPtr);
    //        locZ = CUser::GetZ(v16);
    //        v17 = CSPointer::operator_indirect(&rullerPtr);
    //        locY = CUser::GetY(v17);
    //        v18 = CSPointer::operator_indirect(&rullerPtr);
    //        locX = CUser::GetX(v18);
    //        v19 = CSPointer::operator_indirect(&rullerPtr);
    //        accountId = CUser::GetAccountID(v19);
    //        charId2 = charId;
    //        logType = ;
    //        g_winlog.Add(LOG_IN, L"%d,%d,%d,,,%d,%d,%d,%s,,,%d,%d,%d,%d,%d,%d,%d,,,,%s,%s,,,", LogId_TryDismissPledge, charId2, accountId, locX, locY, locZ, pledgeName2, race, gender, class_1, level, 0, pledgeId2, 0, charName, accountName);
    //        unguard();
    //        return true;
    //    }

    //    if (updateValue == 0)
    //    {
    //        v20 = CSPointer::operator_indirect(&rullerPtr)->m_d.m_charId_3;
    //        v21 = CSPointer::operator_indirect(&rullerPtr);
    //        accountName = CUser::GetAccountName(v21);
    //        v22 = CSPointer::operator_indirect(&rullerPtr);
    //        charName = CUser::GetCharName(v22);
    //        pledgeId2 = pledgeId;
    //        v23 = CSPointer::operator_indirect(&rullerPtr);
    //        level = CUser::GetLevel(v23);
    //        v24 = CSPointer::operator_indirect(&rullerPtr);
    //        class_1 = CUser::GetClass(v24);
    //        v25 = CSPointer::operator_indirect(&rullerPtr);
    //        gender = CUser::GetGender(v25);
    //        v26 = CSPointer::operator_indirect(&rullerPtr);
    //        race = CUser::GetRace(v26);
    //        pledgeName2 = pledgeName;
    //        v27 = CSPointer::operator_indirect(&rullerPtr);
    //        locZ = CUser::GetZ(v27);
    //        v28 = CSPointer::operator_indirect(&rullerPtr);
    //        locY = CUser::GetY(v28);
    //        v29 = CSPointer::operator_indirect(&rullerPtr);
    //        locX = CUser::GetX(v29);
    //        v30 = CSPointer::operator_indirect(&rullerPtr);
    //        accountId = CUser::GetAccountID(v30);
    //        charId2 = v20;

    //        g_winlog.Add(LOG_IN, L"%d,%d,%d,,,%d,%d,%d,%s,,,%d,%d,%d,%d,%d,%d,%d,,,,%s,%s,,,", LogId_UndoDismissPledge, charId2, accountId, locX, locY, locZ, pledgeName2, race, gender, class_1, level, 0, pledgeId2, 0, charName, accountName);
    //        unguard();
    //        return true;
    //    }

    unguard();
    return true;
}

// L2CacheD 0x0042F840
bool CPledgeDb::ModifyPledgeName(uint32_t pledgeId, const wchar_t* newName)
{
    guard(L"bool CPledgeDb::ModifyPledgeName(uint32_t nPledgeId, wchar_t *sNewName)", L"bool CPledgeDb::ModifyPledgeName(uint32_t nPledgeId, wchar_t *sNewName) (exit)", TEXT(__FILE__), __LINE__);

    CPledgeSP pledge = GetPledge(pledgeId);

    if (pledge == NULL)
    {
        unguard();
        return false;
    }

    wchar_t pledgeName[26];
    ::memset(pledgeName, 0, sizeof(pledgeName));

    wchar_t pledgeNameLower[26];
    ::memset(pledgeNameLower, 0, sizeof(pledgeNameLower));

    wchar_t newNameLower[26];
    ::memset(newNameLower, 0, sizeof(newNameLower));

    ::wcscpy(pledgeName, pledge->name());
    Utils::CopyStrInLowerCase(pledgeName, pledgeNameLower);
    Utils::CopyStrInLowerCase(newName, newNameLower);

    DBConn sql;
    if (!sql.Execute(L"EXEC lin_ModifyPledgeName N'%s', %d ", newNameLower, pledgeId))
    {
        unguard();
        return false;
    }

    m_lock.WriteLock();

    pledge->setName(newName);

    m_pledgesById.DeleteObject(pledgeId, NULL, -1);
    m_pledgesByName.DeleteObject(std::wstring(pledgeNameLower), NULL, -1);

    m_pledgesById.Insert(pledgeId, pledge.get(), NULL, -1);
    m_pledgesByName.Insert(std::wstring(pledgeNameLower), pledge.get(), NULL, -1);

    m_lock.WriteUnlock();

    unguard();
    return true;
}

// L2CacheD 0x0042C5B0
void CPledgeDb::LoadPledgeMember(CPledge* pledge)
{
    guard(L"void CPledgeDb::LoadPledgeMember(CPledge* pPledge)", L"int CPledgeDb::LoadPledgeMember(CPledge* pPledge) (exit)", TEXT(__FILE__), __LINE__);

    uint32_t charId = 0;
    DBConn sql;
    sql.Bind(&charId);
    if (sql.Execute(L"EXEC lin_LoadPledgeMember %d", pledge->id()))
    {
        while (sql.Fetch())
        {
            pledge->AddMember(charId);
        }
    }
    else
    {
        g_winlog.Add(LOG_ERROR, L"Error in Loading Pledge[%d] Member", pledge->id());
    }

    unguard();
}

// L2CacheD 0x00430170
void CPledgeDb::JoinPledgeMember(uint32_t pledgeId, uint32_t userId)
{
    guard(L"bool CPledgeDb::JoinPledgeMember(uint32_t nPledgeDbId, uint32_t nUserDbId)", L"bool CPledgeDb::JoinPledgeMember(uint32_t nPledgeDbId, uint32_t nUserDbId) (exit)", TEXT(__FILE__), __LINE__);

    int success = 0;
    DBConn sql;
    sql.Bind(&success);
    if (sql.Execute(L"EXEC lin_JoinPledgeMember %d, %d", pledgeId, userId))
    {
        sql.Fetch();
    }

    if (success == 0)
    {
        unguard();
        return;
    }

    CPledgeSP pledge = GetPledge(pledgeId);
    if (pledge != NULL)
    {
        pledge->AddMember(userId);
    }

    //    CUserSP user = g_userDb.GetUser(userId, true); TODO: port me
    //    if (user != NULL)
    //    {
    //        user->SetPledge(pledgeId);
    //    }

    unguard();
}

// L2CacheD 0x00430350
bool CPledgeDb::WithdrawPledgeMember(uint32_t pledgeId, uint32_t userId)
{
    guard(L"bool CPledgeDb::WithdrawPledgeMember(uint32_t nPledgeDbId, uint32_t nUserDbId)", L"bool CPledgeDb::WithdrawPledgeMember(uint32_t nPledgeDbId, uint32_t nUserDbId) (exit)", TEXT(__FILE__), __LINE__);

    int success = 1;
    DBConn sql;
    sql.Bind(&success);
    if (sql.Execute(L"EXEC lin_WithdrawPledgeMember %d, %d", pledgeId, userId))
    {
        sql.Fetch();
    }

    if (success == 0)
    {
        unguard();
        return false;
    }

    CPledgeSP pledge = GetPledge(pledgeId);
    if (pledge != NULL)
    {
        pledge->RemoveMember(userId);
    }

    //    CUserSP user = g_userDb.GetUser(userId, true); TODO: port me
    //    if (user != NULL)
    //    {
    //        user->SetPledge(0);
    //    }

    unguard();
    return true;
}
