#include "threads/CRWLock.h"

#include "ui/CLog.h"
#include "utils/dbg/StackGuard.h"
#include "utils/dbg/LockStack.h"
#include "threads/ThreadLocalStorage.h"

// L2CacheD 0x004346B0
CRWLock::CRWLock()
    : m_critSection()
    , m_semaphore(0)
    , m_readersCount(0)
{
    guard(L"RWLock::RWLock()", L"RWLock::RWLock() (exit)", TEXT(__FILE__), __LINE__);

    const int INITIAL_COUNT = 1;
    const int MAX_COUNT = 1;
    m_semaphore = ::CreateSemaphoreA(NULL, INITIAL_COUNT, MAX_COUNT, NULL);
    ::InitializeCriticalSection(&m_critSection);

    unguard();
}

// L2CacheD 0x00434700
CRWLock::~CRWLock()
{
    guard(L"RWLock::~RWLock()", L"RWLock::~RWLock() (exit)", TEXT(__FILE__), __LINE__);

    ::CloseHandle(m_semaphore);
    ::DeleteCriticalSection(&m_critSection);

    unguard();
}

// L2CacheD 0x004347B0
void CRWLock::ReadLock()
{
    guard(L"void RWLock::ReadLock()", L"void RWLock::ReadLock() (exit)", TEXT(__FILE__), __LINE__);

    LockStack& lockStack = g_stacks[tls::GetCurrentThreadIndex()];
    lockStack.Push(0, 0, L"RW ReadLock", reinterpret_cast<int>(this));

    ::EnterCriticalSection(&m_critSection);
    if (++m_readersCount == 1)  // was 0, so first read lock
    {
        //  --semaphore
        if (::WaitForSingleObject(m_semaphore, INFINITE) != WAIT_OBJECT_0)  // wait for write lock released
        {
            g_winlog.Add_Ansi(LOG_ERROR, "ReadLock failed on dataLock");
        }
    }
    ::LeaveCriticalSection(&m_critSection);

    unguard();
}

// L2CacheD 0x00434990
void CRWLock::ReadUnlock()
{
    guard(L"void RWLock::ReadUnlock()", L"void RWLock::ReadUnlock() (exit)", TEXT(__FILE__), __LINE__);

    LockStack& lockStack = g_stacks[tls::GetCurrentThreadIndex()];
    lockStack.Pop(0, 0, L"RW ReadUnlock", reinterpret_cast<int>(this));

    ::EnterCriticalSection(&m_critSection);
    --m_readersCount;
    if (m_readersCount == 0)
    {
        long previousCount;
        ::ReleaseSemaphore(m_semaphore, 1, &previousCount);  // ++semaphore
    }
    ::LeaveCriticalSection(&m_critSection);

    unguard();
}

// L2CacheD 0x00434880
void CRWLock::WriteLock()
{
    guard(L"void RWLock::WriteLock()", L"void RWLock::WriteLock() (exit)", TEXT(__FILE__), __LINE__);

    LockStack& lockStack = g_stacks[tls::GetCurrentThreadIndex()];
    lockStack.Push(0, 0, L"RW WriteLock", reinterpret_cast<int>(this));

    ::WaitForSingleObject(m_semaphore, INFINITE);  //  --semaphore

    unguard();
}

// L2CacheD 0x00434A50
void CRWLock::WriteUnlock()
{
    guard(L"void RWLock::WriteUnlock()", L"void RWLock::WriteUnlock() (exit)", TEXT(__FILE__), __LINE__);

    LockStack& lockStack = g_stacks[tls::GetCurrentThreadIndex()];
    lockStack.Pop(0, 0, L"RW WriteUnlock", reinterpret_cast<int>(this));

    long previousCount;
    ::ReleaseSemaphore(m_semaphore, 1, &previousCount);  // ++semaphore
    if (previousCount != 0)
    {
        g_winlog.Add(LOG_ERROR, L"WriteUnlock semaphore was not locked");
    }

    unguard();
}
