#pragma once

#include "Orts.h"
#include "ui/CLog.h"
#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"

template <typename T>
class CSPointer
{
public:
    CSPointer()
        : m_object(NULL)
        , m_workType()
    {
        guard(L"CSPointer()", L"CSPointer() (exit)", TEXT(__FILE__), __LINE__);
        unguard();
    }

    // L2CacheD 0x0040B7D0
    CSPointer(const CSPointer& other)
        : m_object(other.m_object)
        , m_workType(other.m_workType)
    {
        guard(L"CSPointer(const CSPointer& other)", L"CSPointer(const CSPointer& other) (exit)", TEXT(__FILE__), __LINE__);

        if (m_object != NULL)
        {
            m_object->AddRef(__FILE__, __LINE__, ORT_CACHED);
        }

        unguard();
    }

    CSPointer(T* object, ORTs orts)
        : m_object(NULL)
        , m_workType()
    {
        guard(L"CSPointer(T* object, ORTs orts)", L"CSPointer(T* object, ORTs orts) (exit)", TEXT(__FILE__), __LINE__);

        if (object != NULL)
        {
            object->AddRef(__FILE__, __LINE__, orts);
        }

        if (m_object != NULL)
        {
            m_object->Release(__FILE__, __LINE__, m_workType, false);
        }

        m_object = object;
        m_workType = orts;

        unguard();
    }

    CSPointer& operator=(const CSPointer& other)
    {
        guard(L"CSPointer(const CSPointer& other)", L"CSPointer(const CSPointer& other) (exit)", TEXT(__FILE__), __LINE__);
        if (this == &other)
        {
            unguard();
            return *this;
        }

        SetObject(other.m_object, other.m_workType);
        unguard();
        return *this;
    }

    // TODO: this is dangerous, but is present in original L2 code. Hide it after porting
    // operator T*()
    T* get()
    {
        guard(L"operator T*()", L"operator T*() (exit)", TEXT(__FILE__), __LINE__);
        T* object = m_object;
        unguard();
        return object;
    }

    // L2CacheD T=CUser 0x004151F0, T=CAlliance/CPledge 0x00405300
    T* operator->() const
    {
        guard(L"T* operator->() const", L"T* operator->() const (exit)", TEXT(__FILE__), __LINE__);

        if (m_object == NULL)
        {
            g_winlog.Add_Ansi(LOG_ERROR, "[%s][%d] referencing NULL object.", __FILE__, __LINE__);
            CallStack::ReportCallStack();
        }

        T* object = m_object;
        unguard();
        return object;
    }

    // TODO: port me after c++11
    bool operator==(void* null) const
    {
        return m_object == null;
    }

    bool operator!=(void* null) const
    {
        return m_object != null;
    }

    void SetObject(T* object, ORTs orts)
    {
        guard(L"void SetObject(const T* p, ORTs orts)", L"void SetObject(const T* p, ORTs orts) (exit)", TEXT(__FILE__), __LINE__);

        if (object != NULL)
        {
            object->AddRef(__FILE__, __LINE__, orts);
        }

        if (m_object != NULL)
        {
            m_object->Release(__FILE__, __LINE__, m_workType, false);
        }

        m_object = object;
        m_workType = orts;

        unguard();
    }

    ~CSPointer()
    {
        if (m_object != NULL)
        {
            m_object->Release(__FILE__, __LINE__, ORT_CACHED, false);
        }
    }

private:
    T* m_object;
    ORTs m_workType;
};
