#include "threads/SpinLock.h"

#include "utils/dbg/StackGuard.h"

SpinLock::SpinLock()
    : m_lock(0)
{
}

void SpinLock::Enter()
{
    if (::InterlockedExchange(&m_lock, 1))
    {
        Wait();
    }
}

void SpinLock::Leave()
{
    ::InterlockedExchange(&m_lock, 0);
}

bool SpinLock::TryEnter()
{
    return ::InterlockedExchange(&m_lock, 1) == 0;
}

// 0x004086E0
void SpinLock::Wait()
{
    guard(L"void SpinLock::Wait()", L"void SpinLock::Wait() (exit)", TEXT(__FILE__), __LINE__);

    int count = 4000;
    while (--count >= 0)
    {
        if (::InterlockedExchange(&m_lock, 1) == 0)
        {
            unguard();
            return;
        }
        _mm_pause();
        --count;
    }

    count = 3999;
    while (true)
    {
        ::SwitchToThread();
        if (::InterlockedExchange(&m_lock, 1) == 0)
        {
            break;
        }

        if (--count < 0)
        {
            do
            {
                Sleep(1000u);
            } while (::InterlockedExchange(&m_lock, 1));

            break;
        }
    }

    unguard();
}
