#include "MemoryObject.h"

#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"
#include "memory/Allocator.h"

#include "ui/CLog.h"

long MemoryObject::s_instances(0);      // 0x01BD4FC4
long MemoryObject::s_nAlloc(0);         // 0x01BD4FC8
long MemoryObject::s_refCountSlots(0);  // 0x01BD3B60

// 0x0040DAB0
MemoryObject::MemoryObject()
    : m_nRefCount(0)
    , m_identity(0)
    , m_inTimer(false)
{
    guard(L"MemoryObject::MemoryObject()");

    m_identity = ::InterlockedIncrement(&MemoryObject::s_instances);
    AddRef(__FILE__, __LINE__, ORT_CREATE_DELETE);

    unguard();
}

// 0x00409210
MemoryObject::~MemoryObject()
{
}

// 0x00409210 inlined
void MemoryObject::operator delete(void* /*ptr*/)
{
    g_winlog.Add(LOG_ERROR, L"Never Call Oprator Delete!!! MemoryObject at file[%s], line[%d]", TEXT(__FILE__), __LINE__);
    CallStack::ReportCallStack();
}

// 0x00408980
long MemoryObject::AddRef(const char* /*fileName*/, int /*line*/, ORTs /*workType*/)
{
    ::InterlockedIncrement(&m_nRefCount);
    return m_nRefCount;
}

// 0x00424CC8
bool MemoryObject::vfunc10() const
{
    return false;
}

// 0x00424CD0
bool MemoryObject::IsObject() const
{
    return false;
}

// 0x004091F0
bool MemoryObject::IsUser() const
{
    return false;
}

// 0x00424CD8
bool MemoryObject::IsItem() const
{
    return false;
}

// 0x00409200
long MemoryObject::RefCount() const
{
    return m_nRefCount;
}

// 0x0040DB00
void MemoryObject::ReportOrt(int a2, int a3)
{
    guard(L"void MemoryObject::ReportOrt()");
    unguard();
}

void MemoryObject::setInTimer(bool isInTimer)
{
    m_inTimer = isInTimer;
}

bool MemoryObject::inTimer() const
{
    return m_inTimer;
}

// 0x0040DB20
void MemoryObject::Release(const char* file, int line, ORTs /*nWorkType*/, bool /*flag*/)
{
    long refCountAfterRelease = ::InterlockedDecrement(&m_nRefCount);

    if (refCountAfterRelease > 0)
    {
        return;
    }

    if (refCountAfterRelease != 0)
    {
        if (refCountAfterRelease > -1)
        {
            return;
        }

        g_winlog.Add(LOG_ERROR, "Invalid Memory (%d) on MemoryObject(%d)", refCountAfterRelease, m_identity);
        if (file)
        {
            g_winlog.Add(LOG_ERROR, "           Caller : file[%s], line[%d]", file, line);
        }

        ReportOrt(0, 0);
        CallStack::ReportCallStack();

        return;
    }

    if (m_inTimer)
    {
        if (file)
        {
            g_winlog.Add(LOG_ERROR, "deleting in-timer-object![MemoryObject] at file[%s], line[%d]", file, line);
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[MemoryObject::s_nAlloc]);
        }
        else
        {
            g_winlog.Add(LOG_ERROR, "deleting in-timer-object![MemoryObject]");
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[MemoryObject::s_nAlloc]);
        }
        return;
    }

    this->~MemoryObject();

    g_allocator.deallocate(this);

    ::InterlockedDecrement(&Allocator::g_allocatedObjects[MemoryObject::s_nAlloc]);
}
