#include "db/DBConn.h"
#include "db/SqlConnection.h"
#include "ui/CLog.h"
#include "utils/dbg/StackGuard.h"
#include "config/Config.h"
#include "ui/resources.h"
#include "AppInstance.h"

#include <Sqlext.h>
#include <Winreg.h>
#include <sql.h>

#include <wchar.h>

SqlConnection* DBConn::s_sqlConnections1(NULL);   // 0x0042E7E8
SqlConnection* DBConn::s_sqlConnections2(NULL);   // 0x0042E7E4
SqlConnection* DBConn::s_sqlConnections3(NULL);   // 0x0042E7E0
wchar_t DBConn::s_connStrBBS[256];                // 0x0042E7F0
wchar_t DBConn::s_connStr[256];                   // 0x0042E9F0
SQLHANDLE DBConn::s_sqlEnvHandle(NULL);           // 0x0042EBF0
SqlConnection* DBConn::s_connectionsHead1(NULL);  // 0x0042EBF4
SqlConnection* DBConn::s_connectionsHead2(NULL);  // 0x0042EBF8
SqlConnection* DBConn::s_connectionsHead3(NULL);  // 0x0042EBFC
CIOCriticalSection DBConn::s_lock1;               // 0x0042EC00
CIOCriticalSection DBConn::s_lock2;               // 0x0042EC1C
int DBConn::s_dbConnectionNum1(16);               // 0x0042D7D4
int DBConn::s_dbConnectionNum2(0);                // 0x019D7DF4
int DBConn::s_dbConnectionNum3(16);               // 0x019D7E20

// 0x00401000
DBConn::DBConn()
    : m_sqlHandler()
    , m_currentCollumn(1)
    , m_sqlConnection(NULL)
    , m_poolIndex(Pool_1)
{
    guard(L"DBConn::DBConn");

    int sleepCount = 0;
    while (true)
    {
        DBConn::s_lock1.Enter(NULL, 0);

        if (DBConn::s_connectionsHead1->nextConnection)
        {
            m_sqlConnection = DBConn::s_connectionsHead1;
            DBConn::s_connectionsHead1 = DBConn::s_connectionsHead1->nextConnection;
        }

        DBConn::s_lock1.Leave(NULL, 0);

        if (m_sqlConnection != NULL)
        {
            break;
        }

        g_winlog.Add(LOG_ERROR, L"DBConn::DBConn no available freeSqlPool. Sleep 1 second. (%d times)", ++sleepCount);
        ::Sleep(1000u);
    }

    m_sqlHandler = m_sqlConnection->stmtSqlHandler;

    memset(m_lastQuery, 0, sizeof(m_lastQuery));

    unguard();
}

// 0x004010F0
DBConn::~DBConn()
{
    guard(L"DBConn::DBConn::~DBConn");

    if (m_poolIndex == Pool_1)
    {
        ::SQLFreeStmt(m_sqlHandler, SQL_HANDLE_DBC);
        ::SQLFreeStmt(m_sqlHandler, SQL_NULL_HANDLE);
        ::SQLFreeStmt(m_sqlHandler, SQL_HANDLE_STMT);

        DBConn::s_lock1.Enter(0, 0);

        m_sqlConnection->nextConnection = DBConn::s_connectionsHead1;
        DBConn::s_connectionsHead1 = m_sqlConnection;

        DBConn::s_lock1.Leave(0, 0);

        unguard();
        return;
    }

    if (m_poolIndex == Pool_BbsDb)
    {
        if (DBConn::s_dbConnectionNum2 > 0)
        {
            ::SQLFreeStmt(m_sqlHandler, SQL_HANDLE_DBC);
            ::SQLFreeStmt(m_sqlHandler, SQL_NULL_HANDLE);
            ::SQLFreeStmt(m_sqlHandler, SQL_HANDLE_STMT);

            DBConn::s_lock2.Enter(0, 0);

            m_sqlConnection->nextConnection = DBConn::s_connectionsHead2;
            DBConn::s_connectionsHead2 = m_sqlConnection;

            DBConn::s_lock2.Leave(0, 0);
        }
        unguard();
        return;
    }

    if (DBConn::s_dbConnectionNum3 > 0)
    {
        ::SQLFreeStmt(m_sqlHandler, SQL_HANDLE_DBC);
        ::SQLFreeStmt(m_sqlHandler, SQL_NULL_HANDLE);
        ::SQLFreeStmt(m_sqlHandler, SQL_HANDLE_STMT);

        DBConn::s_lock2.Enter(0, 0);

        m_sqlConnection->nextConnection = DBConn::s_connectionsHead3;
        DBConn::s_connectionsHead3 = m_sqlConnection;

        DBConn::s_lock2.Leave(0, 0);
    }

    unguard();
}

// 0x00402340
void DBConn::Init()
{
    guard(L"DBConn::DBConn::Init");

    ::SQLSetEnvAttr((SQLHENV)SQL_NULL_HANDLE, SQL_ATTR_CONNECTION_POOLING, (SQLPOINTER)SQL_CP_ONE_PER_DRIVER, SQL_TINYINT);

    DBConn::s_sqlEnvHandle = SQL_NULL_HANDLE;
    SQLRETURN allocRes = ::SQLAllocHandle(SQL_HANDLE_ENV, (SQLHANDLE)SQL_NULL_HANDLE, &DBConn::s_sqlEnvHandle);
    if (allocRes != SQL_SUCCESS && allocRes != SQL_SUCCESS_WITH_INFO || (allocRes = ::SQLSetEnvAttr(DBConn::s_sqlEnvHandle, SQL_ATTR_ODBC_VERSION, (SQLPOINTER)SQL_OV_ODBC3, 0)) != SQL_SUCCESS && allocRes != SQL_SUCCESS_WITH_INFO)
    {
        g_winlog.Add(LOG_ERROR, L"db env allocation failed");
        unguard();
        return;
    }

    if (!DBConn::Login())
    {
        g_winlog.Add(LOG_ERROR, L"db login failed");
        ::SQLFreeHandle(SQL_HANDLE_ENV, DBConn::s_sqlEnvHandle);
        DBConn::s_sqlEnvHandle = SQL_NULL_HANDLE;
        unguard();
        return;
    }

    DBConn::AllocSQLPool();
    DBConn::AllocSQLPool2();

    if (Config::s_dbPair > 0)
    {
        DBConn::AllocSQLPool3();
    }

    unguard();
}

// 0x00401FC0
bool DBConn::Fetch()
{
    guard(L"DBConn::Fetch");

    if (m_sqlHandler == SQL_NULL_HANDLE)
    {
        g_winlog.Add(LOG_ERROR, L"Fetch error = (STMT)");
        unguard();
        return false;
    }

    SQLRETURN sqlResult = ::SQLFetch(m_sqlHandler);
    if (sqlResult != SQL_SUCCESS && sqlResult != SQL_SUCCESS_WITH_INFO)
    {
        if (sqlResult != SQL_NO_DATA)
        {
            SQLSMALLINT pcchErrorMsg;
            SQLINTEGER pfNativeError;
            wchar_t szSqlState[QUERY_MAX_LEN];
            wchar_t szErrorMsg[QUERY_MAX_LEN];

            ::SQLGetDiagRecW(SQL_HANDLE_STMT, m_sqlHandler, SQL_HANDLE_ENV, szSqlState, &pfNativeError, szErrorMsg, QUERY_MAX_LEN, &pcchErrorMsg);
            g_winlog.Add(LOG_ERROR, L"fetch error");
            g_winlog.Add(LOG_ERROR, L"sqlstate %s", szSqlState);
            g_winlog.Add(LOG_ERROR, L"msgText %s", szErrorMsg);
            g_winlog.Add(LOG_ERROR, L"sql %s", m_lastQuery);
        }
        unguard();
        return false;
    }

    unguard();
    return true;
}

// 0x00402250
void DBConn::Bind(wchar_t* value, int length)
{
    guard(L"DBConn::Bind");

    SQLUSMALLINT curCollumn = m_currentCollumn;
    ++m_currentCollumn;
    ::SQLBindCol(m_sqlHandler, curCollumn, SQL_C_WCHAR, value, length, NULL);

    unguard();
}

// 0x004022F0
void DBConn::BindNullable(int* value, SQLINTEGER* indicator)
{
    guard(L"void DBConn::BindNullable(int *pValue, SQLLEN *pIndicator)");

    SQLUSMALLINT curCollumn = m_currentCollumn;
    ++m_currentCollumn;
    ::SQLBindCol(m_sqlHandler, curCollumn, SQL_INTEGER, value, 4, indicator);

    unguard();
}

// 0x004022A0
void DBConn::ResetHtmt()
{
    // guard(L"DBConn::Bind");  // FIXED
    guard(L"DBConn::ResetHtmt");

    m_currentCollumn = 1;

    ::SQLFreeStmt(m_sqlHandler, SQL_UNBIND);
    ::SQLFreeStmt(m_sqlHandler, SQL_HANDLE_STMT);
    ::SQLCloseCursor(m_sqlHandler);

    unguard();
}

// 0x00402450
bool DBConn::Execute(const wchar_t* format, ...)
{
    va_list va;
    va_start(va, format);

    guard(L"DBConn::Execute");

    if ((m_poolIndex == Pool_BbsDb) && (DBConn::s_dbConnectionNum2 == 0))
    {
        unguard();
        return false;
    }

    wchar_t sqlQuery[0x2000];
    ::memset(sqlQuery, 0, sizeof(sqlQuery));  // FIXED
    // int len = vswprintf(sqlQuery, format, va);  // FIXED
    int len = vswprintf(sqlQuery, sizeof(sqlQuery) / sizeof(*sqlQuery), format, va);
    int queryLen = wcslen(sqlQuery);
    int maxLen = min(queryLen, QUERY_MAX_LEN - 1);

    SQLRETURN sqlResult = SQL_ERROR;
    if (len > 0)
    {
        DWORD before = ::GetTickCount();
        sqlResult = ::SQLExecDirectW(m_sqlHandler, sqlQuery, len);
        DWORD duration = ::GetTickCount() - before;

        if (duration > Config::s_sqlExecLimit)
        {
            wchar_t logMsg[QUERY_MAX_LEN];
            memset(logMsg, 0, sizeof(logMsg));

            wcsncpy(logMsg, sqlQuery, maxLen);
            g_winlog.Add(LOG_ERROR, L"sql execution time [%d]ms, sql[%s]", duration, logMsg);
        }
    }

    wcsncpy(m_lastQuery, sqlQuery, maxLen);
    if (sqlResult != SQL_SUCCESS && sqlResult != SQL_SUCCESS_WITH_INFO)
    {
        DBConn::Error(SQL_HANDLE_STMT, m_sqlHandler, sqlQuery);
        unguard();
        return false;
    }

    unguard();
    return true;
}

// 0x00401250
void DBConn::AllocSQLPool()
{
    guard(L"DBConn::DBConn::AllocSQLPool");

    DBConn::s_sqlConnections1 = new SqlConnection[DBConn::s_dbConnectionNum1];
    for (int i = 0; i < DBConn::s_dbConnectionNum1; ++i)
    {
        DBConn::s_sqlConnections1[i].nextConnection = NULL;
        SQLRETURN sqlResult = ::SQLAllocHandle(SQL_HANDLE_DBC, DBConn::s_sqlEnvHandle, &DBConn::s_sqlConnections1[i].dbcSqlHandler);

        if ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            g_winlog.Add(LOG_ERROR, L"hdbc allocation failed");
            DBConn::s_sqlConnections1[i].stmtSqlHandler = NULL;
            DBConn::s_sqlConnections1[i].dbcSqlHandler = NULL;
            continue;
        }

        ::SQLSetConnectAttrW(DBConn::s_sqlConnections1[i].dbcSqlHandler, SQL_ATTR_LOGIN_TIMEOUT, (SQLPOINTER)60, NULL);
        ::SQLSetConnectAttrW(DBConn::s_sqlConnections1[i].dbcSqlHandler, SQL_ATTR_CONNECTION_TIMEOUT, (SQLPOINTER)60, NULL);

        wchar_t connStrOut[1024];
        memset(connStrOut, 0, sizeof(connStrOut));
        int outLen = lstrlenW(connStrOut);
        int len = lstrlenW(DBConn::s_connStr);
        SQLSMALLINT pcchConnStrOut;

        sqlResult = ::SQLDriverConnectW(DBConn::s_sqlConnections1[i].dbcSqlHandler, NULL, DBConn::s_connStr, len, connStrOut, outLen, &pcchConnStrOut, NULL);
        if ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            g_winlog.Add(LOG_ERROR, L"hdbc connection failed");
            continue;
        }

        sqlResult = ::SQLAllocHandle(SQL_HANDLE_STMT, DBConn::s_sqlConnections1[i].dbcSqlHandler, &DBConn::s_sqlConnections1[i].stmtSqlHandler);
        if ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            g_winlog.Add(LOG_ERROR, L"stmt allocation failed");
            ::SQLFreeHandle(SQL_HANDLE_DBC, DBConn::s_sqlConnections1[i].dbcSqlHandler);
            DBConn::s_sqlConnections1[i].stmtSqlHandler = NULL;
            DBConn::s_sqlConnections1[i].dbcSqlHandler = NULL;
        }
    }

    for (int i = 0; i < DBConn::s_dbConnectionNum1; ++i)
    {
        if (DBConn::s_sqlConnections1[i].stmtSqlHandler == SQL_NULL_HANDLE)
        {
            continue;
        }

        if (DBConn::s_connectionsHead1 != NULL)
        {
            DBConn::s_sqlConnections1[i].nextConnection = DBConn::s_connectionsHead1;
            DBConn::s_connectionsHead1 = &DBConn::s_sqlConnections1[i];
        }
        else
        {
            DBConn::s_connectionsHead1 = &DBConn::s_sqlConnections1[i];
        }
    }

    unguard();
}

// 0x00401470
void DBConn::AllocSQLPool2()
{
    guard(L"DBConn::DBConn::AllocSQLPool2");

    if (DBConn::s_dbConnectionNum2 <= 0)
    {
        unguard();
        return;
    }

    DBConn::s_sqlConnections2 = new SqlConnection[DBConn::s_dbConnectionNum2];

    for (int i = 0; i < DBConn::s_dbConnectionNum2; ++i)
    {
        DBConn::s_sqlConnections2[i].nextConnection = NULL;
        SQLRETURN sqlResult = ::SQLAllocHandle(SQL_HANDLE_DBC, DBConn::s_sqlEnvHandle, &DBConn::s_sqlConnections2[i].dbcSqlHandler);

        if ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            // g_winlog.Add(LOG_ERROR, L"hdbc allocation failed"); // FIXED
            g_winlog.Add(LOG_ERROR, L"hdbc bbs allocation failed");
            DBConn::s_sqlConnections2[i].stmtSqlHandler = NULL;
            DBConn::s_sqlConnections2[i].dbcSqlHandler = NULL;
            continue;
        }

        ::SQLSetConnectAttrW(DBConn::s_sqlConnections2[i].dbcSqlHandler, SQL_ATTR_LOGIN_TIMEOUT, (SQLPOINTER)60, NULL);
        ::SQLSetConnectAttrW(DBConn::s_sqlConnections2[i].dbcSqlHandler, SQL_ATTR_CONNECTION_TIMEOUT, (SQLPOINTER)60, NULL);

        wchar_t connStrOut[1024];
        memset(connStrOut, 0, sizeof(connStrOut));
        int outLen = lstrlenW(connStrOut);
        int len = lstrlenW(DBConn::s_connStrBBS);
        SQLSMALLINT pcchConnStrOut;

        sqlResult = ::SQLDriverConnectW(DBConn::s_sqlConnections2[i].dbcSqlHandler, NULL, DBConn::s_connStrBBS, len, connStrOut, outLen, &pcchConnStrOut, NULL);
        if ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            // g_winlog.Add(LOG_ERROR, L"hdbc connection failed"); // FIXED
            g_winlog.Add(LOG_ERROR, L"hdbc bbs connection failed");
            continue;
        }

        sqlResult = ::SQLAllocHandle(SQL_HANDLE_STMT, DBConn::s_sqlConnections2[i].dbcSqlHandler, &DBConn::s_sqlConnections2[i].stmtSqlHandler);
        if ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            // g_winlog.Add(LOG_ERROR, L"stmt allocation failed"); // FIXED
            g_winlog.Add(LOG_ERROR, L"stmt bbs allocation failed");

            ::SQLFreeHandle(SQL_HANDLE_DBC, DBConn::s_sqlConnections2[i].dbcSqlHandler);
            DBConn::s_sqlConnections2[i].stmtSqlHandler = NULL;
            DBConn::s_sqlConnections2[i].dbcSqlHandler = NULL;
        }
    }

    for (int i = 0; i < DBConn::s_dbConnectionNum2; ++i)
    {
        if (DBConn::s_sqlConnections2[i].stmtSqlHandler == SQL_NULL_HANDLE)
        {
            continue;
        }

        ::SQLSetStmtAttrW(DBConn::s_sqlConnections2[i].stmtSqlHandler, SQL_ATTR_QUERY_TIMEOUT, (SQLPOINTER)6, NULL);
        if (DBConn::s_connectionsHead2)
        {
            DBConn::s_sqlConnections2[i].nextConnection = DBConn::s_connectionsHead2;
            DBConn::s_connectionsHead2 = &DBConn::s_sqlConnections2[i];
        }
        else
        {
            DBConn::s_connectionsHead2 = &DBConn::s_sqlConnections2[i];
        }
    }

    unguard();
}

// 0x004016C0
void DBConn::AllocSQLPool3()
{
    guard(L"DBConn::DBConn::AllocSQLPool3");

    // if (DBConn::s_dbConnectionNum2 <= 0) // FIXED
    if (DBConn::s_dbConnectionNum3 <= 0)
    {
        unguard();
        return;
    }

    DBConn::s_sqlConnections3 = new SqlConnection[DBConn::s_dbConnectionNum3];
    for (int i = 0; i < DBConn::s_dbConnectionNum3; ++i)
    {
        DBConn::s_sqlConnections3[i].nextConnection = NULL;
        SQLRETURN sqlResult = ::SQLAllocHandle(SQL_HANDLE_DBC, DBConn::s_sqlEnvHandle, &DBConn::s_sqlConnections3[i].dbcSqlHandler);

        if ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            // g_winlog.Add(LOG_ERROR, L"hdbc allocation failed"); // FIXED
            g_winlog.Add(LOG_ERROR, L"hdbc bbs2 allocation failed");
            DBConn::s_sqlConnections3[i].stmtSqlHandler = 0;
            DBConn::s_sqlConnections3[i].dbcSqlHandler = 0;
            continue;
        }

        ::SQLSetConnectAttrW(DBConn::s_sqlConnections3[i].dbcSqlHandler, SQL_ATTR_LOGIN_TIMEOUT, (SQLPOINTER)60, 0);
        ::SQLSetConnectAttrW(DBConn::s_sqlConnections3[i].dbcSqlHandler, SQL_ATTR_CONNECTION_TIMEOUT, (SQLPOINTER)60, 0);

        wchar_t connStrOut[1024];
        memset(connStrOut, 0, sizeof(connStrOut));
        int outLen = lstrlenW(connStrOut);
        int len = lstrlenW(DBConn::s_connStrBBS);
        SQLSMALLINT pcchConnStrOut;

        sqlResult = ::SQLDriverConnectW(DBConn::s_sqlConnections3[i].dbcSqlHandler, 0, DBConn::s_connStrBBS, len, connStrOut, outLen, &pcchConnStrOut, 0);
        if ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            // g_winlog.Add(LOG_ERROR, L"hdbc connection failed"); // FIXED
            g_winlog.Add(LOG_ERROR, L"hdbc bbs2 connection failed");
            continue;
        }

        sqlResult = ::SQLAllocHandle(SQL_HANDLE_STMT, DBConn::s_sqlConnections3[i].dbcSqlHandler, &DBConn::s_sqlConnections3[i].stmtSqlHandler);
        if ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            // g_winlog.Add(LOG_ERROR, L"stmt allocation failed"); // FIXED
            g_winlog.Add(LOG_ERROR, L"stmt bbs2 allocation failed");

            ::SQLFreeHandle(SQL_HANDLE_DBC, DBConn::s_sqlConnections3[i].dbcSqlHandler);
            DBConn::s_sqlConnections3[i].stmtSqlHandler = 0;
            DBConn::s_sqlConnections3[i].dbcSqlHandler = 0;
        }
    }

    for (int i = 0; i < DBConn::s_dbConnectionNum3; ++i)
    {
        if (DBConn::s_sqlConnections3[i].stmtSqlHandler == SQL_NULL_HANDLE)
        {
            continue;
        }

        ::SQLSetStmtAttrW(DBConn::s_sqlConnections3[i].stmtSqlHandler, SQL_ATTR_QUERY_TIMEOUT, (SQLPOINTER)6, NULL);
        if (DBConn::s_connectionsHead3)
        {
            DBConn::s_sqlConnections3[i].nextConnection = DBConn::s_connectionsHead3;
            DBConn::s_connectionsHead3 = &DBConn::s_sqlConnections3[i];
        }
        else
        {
            DBConn::s_connectionsHead3 = &DBConn::s_sqlConnections3[i];
        }
    }

    unguard();
}

// 0x00401910
BOOL DBConn::loginDlgProc(HWND hDlg, UINT msgType, WPARAM wparam, LPARAM)
{
    guard(L"DBConn::DBConn::loginDlgProc");

    if (msgType == WM_INITDIALOG)
    {
        ::SendDlgItemMessageW(hDlg, CONTROL_LOGDB_LOGIN_EDIT_BOX_FILE_DB, WM_SETTEXT, 0, (LPARAM)L"linworld");
    }
    else if (msgType == WM_COMMAND && wparam == IDOK)
    {
        wchar_t buffer[68];

        ::SendDlgItemMessageW(hDlg, CONTROL_LOGDB_LOGIN_EDIT_BOX_FILE_DB, WM_GETTEXT, 64u, (LPARAM)buffer);
        lstrcpyW(DBConn::s_connStr, L"FILEDSN=");
        lstrcatW(DBConn::s_connStr, buffer);

        ::SendDlgItemMessageW(hDlg, CONTROL_LOGDB_LOGIN_EDIT_BOX_LOGIN_NAME, WM_GETTEXT, 64u, (LPARAM)buffer);
        lstrcatW(DBConn::s_connStr, L";UID=");
        lstrcatW(DBConn::s_connStr, buffer);

        ::SendDlgItemMessageW(hDlg, CONTROL_LOGDB_LOGIN_EDIT_BOX_PASSWORD, WM_GETTEXT, 64u, (LPARAM)buffer);
        lstrcatW(DBConn::s_connStr, L";PWD=");
        lstrcatW(DBConn::s_connStr, buffer);

        SaveStrToReg(DBConn::s_connStr, L"connStr", L"Software\\PROJECT_L2\\L2LOGD");

        ::EndDialog(hDlg, 0);
    }

    unguard();

    return FALSE;
}

// 0x00401A70
BOOL DBConn::loginBBSDlgProc(HWND hDlg, UINT msgType, WPARAM wparam, LPARAM)
{
    guard(L"DBConn::DBConn::loginBBSDlgProc");

    if (msgType == WM_INITDIALOG)
    {
        ::SendDlgItemMessageW(hDlg, CONTROL_BBS_LOGIN_EDIT_BOX_FILE_DB, WM_SETTEXT, 0, (LPARAM)L"linworld");
    }
    else if (msgType == WM_COMMAND && wparam == IDOK)
    {
        wchar_t buffer[68];

        ::SendDlgItemMessageW(hDlg, CONTROL_BBS_LOGIN_EDIT_BOX_FILE_DB, WM_GETTEXT, 64u, (LPARAM)buffer);
        lstrcpyW(DBConn::s_connStrBBS, L"FILEDSN=");
        lstrcatW(DBConn::s_connStrBBS, buffer);

        ::SendDlgItemMessageW(hDlg, CONTROL_BBS_LOGIN_EDIT_BOX_LOGIN_NAME, WM_GETTEXT, 64u, (LPARAM)buffer);
        lstrcatW(DBConn::s_connStrBBS, L";UID=");
        lstrcatW(DBConn::s_connStrBBS, buffer);

        ::SendDlgItemMessageW(hDlg, CONTROL_BBS_LOGIN_EDIT_BOX_PASSWORD, WM_GETTEXT, 64u, (LPARAM)buffer);
        lstrcatW(DBConn::s_connStrBBS, L";PWD=");
        lstrcatW(DBConn::s_connStrBBS, buffer);

        SaveStrToReg(DBConn::s_connStrBBS, L"connStrBBS", L"Software\\PROJECT_L2\\L2LOGD");

        ::EndDialog(hDlg, 0);
    }

    unguard();

    return FALSE;
}

// 0x00401BD0
bool DBConn::Login()
{
    guard(L"DBConn::Login");

    if (!DBConn::LoadStrFromReg(DBConn::s_connStr, L"connStr", L"Software\\PROJECT_L2\\L2LOGD"))
    {
        ::DialogBoxParamW(g_instance, MAKEINTRESOURCEW(DIALOG_LOGDB_LOGIN), NULL, &DBConn::loginDlgProc, 0);
    }

    if (Config::s_differentBBSConn)
    {
        if (!DBConn::LoadStrFromReg(DBConn::s_connStrBBS, L"connStrBBS", L"Software\\PROJECT_L2\\L2LOGD"))
        {
            ::DialogBoxParamW(g_instance, MAKEINTRESOURCEW(DIALOG_BBS_LOGIN), NULL, &DBConn::loginBBSDlgProc, 0);
        }
    }
    else
    {
        wcscpy(DBConn::s_connStrBBS, DBConn::s_connStr);
    }

    SQLHANDLE hdbc;
    SQLRETURN allocRes = ::SQLAllocHandle(SQL_HANDLE_DBC, DBConn::s_sqlEnvHandle, &hdbc);
    if ((allocRes != SQL_SUCCESS) && (allocRes != SQL_SUCCESS_WITH_INFO))
    {
        // g_winlog.Add(LOG_ERROR, L"hdbc allocation failed"); FIXED
        g_winlog.Add(LOG_ERROR, L"hdbc env allocation failed");
        unguard();
        return false;
    }

    ::SQLSetConnectAttrW(hdbc, SQL_ATTR_LOGIN_TIMEOUT, (SQLPOINTER)60, 0);
    ::SQLSetConnectAttrW(hdbc, SQL_ATTR_CONNECTION_TIMEOUT, (SQLPOINTER)60, 0);

    wchar_t connStrOut[1024];
    memset(connStrOut, 0, sizeof(connStrOut));
    int outLen = lstrlenW(connStrOut);
    int len = lstrlenW(DBConn::s_connStr);
    SQLSMALLINT textLength;

    SQLRETURN sqlResult = ::SQLDriverConnectW(hdbc, 0, DBConn::s_connStr, len, connStrOut, outLen, &textLength, 0);
    while ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
    {
        wchar_t sqlState[QUERY_MAX_LEN];
        wchar_t errorText[QUERY_MAX_LEN];
        SQLSMALLINT pcchErrorMsg;
        SQLINTEGER nativeError;

        ::SQLGetDiagRecW(SQL_HANDLE_DBC, hdbc, 1, sqlState, &nativeError, errorText, QUERY_MAX_LEN, &pcchErrorMsg);

        g_winlog.Add(LOG_ERROR, L"%s", sqlState);
        g_winlog.Add(LOG_ERROR, L"%s", errorText);

        ::DialogBoxParamW(g_instance, MAKEINTRESOURCEW(DIALOG_LOGDB_LOGIN), NULL, &DBConn::loginDlgProc, 0);

        outLen = lstrlenW(connStrOut);
        len = lstrlenW(DBConn::s_connStr);
    }

    ::SQLDisconnect(hdbc);

    if (Config::s_differentBBSConn == true)
    {
        ::SQLSetConnectAttrW(hdbc, SQL_ATTR_LOGIN_TIMEOUT, (SQLPOINTER)60, 0);
        ::SQLSetConnectAttrW(hdbc, SQL_ATTR_CONNECTION_TIMEOUT, (SQLPOINTER)60, 0);

        wchar_t connStrOut[1024];
        memset(connStrOut, 0, sizeof(connStrOut));
        int outLen = lstrlenW(connStrOut);
        int len = lstrlenW(DBConn::s_connStrBBS);
        SQLSMALLINT textLength;

        SQLRETURN sqlResult = ::SQLDriverConnectW(hdbc, 0, DBConn::s_connStrBBS, len, connStrOut, outLen, &textLength, 0);
        while ((sqlResult != SQL_SUCCESS) && (sqlResult != SQL_SUCCESS_WITH_INFO))
        {
            wchar_t sqlState[QUERY_MAX_LEN];
            wchar_t errorText[QUERY_MAX_LEN];
            SQLSMALLINT pcchErrorMsg;
            SQLINTEGER nativeError;

            ::SQLGetDiagRecW(SQL_HANDLE_DBC, hdbc, 1, sqlState, &nativeError, errorText, QUERY_MAX_LEN, &pcchErrorMsg);

            g_winlog.Add(LOG_ERROR, L"%s", sqlState);
            g_winlog.Add(LOG_ERROR, L"%s", errorText);

            ::DialogBoxParamW(g_instance, MAKEINTRESOURCEW(DIALOG_BBS_LOGIN), NULL, &DBConn::loginBBSDlgProc, 0);
            outLen = lstrlenW(connStrOut);
            len = lstrlenW(DBConn::s_connStrBBS);
        }

        ::SQLDisconnect(hdbc);
    }

    ::SQLFreeHandle(SQL_HANDLE_DBC, hdbc);
    unguard();

    return true;
}

// 0x00402120
void DBConn::Error(SQLSMALLINT handleType, SQLHANDLE handle, const wchar_t* query)
{
    guard(L"DBConn::Error");

    SQLSMALLINT pcchErrorMsg;
    SQLINTEGER pfNativeError;
    wchar_t szErrorMsg[QUERY_MAX_LEN];
    wchar_t szSqlState[QUERY_MAX_LEN];

    SQLRETURN sqlResult = ::SQLGetDiagRecW(handleType, handle, 1, szSqlState, &pfNativeError, szErrorMsg, 1024, &pcchErrorMsg);
    if (query != NULL)
    {
        wchar_t truncatedQuery[QUERY_MAX_LEN];
        memset(truncatedQuery, 0, sizeof(truncatedQuery));

        int queryLen = wcslen(query);
        int maxLen = max(queryLen, QUERY_MAX_LEN - 1);

        wcsncpy(truncatedQuery, query, maxLen);
        g_winlog.Add(LOG_ERROR, L"Error sql: %s (%d)", truncatedQuery, sqlResult);
    }

    if (sqlResult == SQL_SUCCESS)
    {
        g_winlog.Add(LOG_ERROR, L"Error %s:%s", szSqlState, szErrorMsg);
    }

    unguard();
}

// 0x004060B0
bool DBConn::LoadStrFromReg(wchar_t* str, const wchar_t* keyStr, const wchar_t* regEntry)
{
    guard(L"bool LoadStrFromReg(wchar_t* str, const wchar_t* keyStr, const wchar_t* regEntry)");

    bool succeed = false;

    HKEY regKey;
    LSTATUS regResult = ::RegOpenKeyExW(HKEY_CURRENT_USER, regEntry, 0, KEY_READ, &regKey);
    if (regResult == ERROR_SUCCESS)
    {
        // DWORD length = 256; fixed
        wchar_t storedValue[256];
        DWORD length = sizeof(storedValue);
        DWORD type;

        regResult = ::RegQueryValueExW(regKey, keyStr, 0, &type, (LPBYTE)storedValue, &length);
        if (regResult == ERROR_SUCCESS && type == REG_BINARY)
        {
            succeed = true;
        }

        ::RegCloseKey(regKey);

        if (succeed)
        {
            lstrcpyW(str, storedValue);
            unguard();
            return true;
        }
    }

    regResult = ::RegOpenKeyExW(HKEY_LOCAL_MACHINE, regEntry, 0, KEY_READ, &regKey);
    if (regResult == ERROR_SUCCESS)
    {
        // DWORD length = 256; fixed
        wchar_t storedValue[256];
        DWORD length = sizeof(storedValue);
        DWORD type;

        regResult = ::RegQueryValueExW(regKey, keyStr, 0, &type, (LPBYTE)storedValue, &length);
        if (regResult == ERROR_SUCCESS && type == REG_BINARY)
        {
            succeed = true;
        }

        ::RegCloseKey(regKey);

        if (succeed)
        {
            lstrcpyW(str, storedValue);
            unguard();
            return true;
        }
    }

    unguard();
    return false;
}

// 0x004061C0
void DBConn::SaveStrToReg(wchar_t* str, const wchar_t* keyStr, const wchar_t* regEntry)
{
    guard(L"void SaveStrToReg(wchar_t* str, const wchar_t* keyStr, const wchar_t* regEntry)");

    wchar_t buffer[256];
    lstrcpyW(buffer, str);

    wchar_t keyClass[4];
    HKEY phkResult;
    DWORD dwDisposition;
    LSTATUS result = ::RegCreateKeyExW(HKEY_LOCAL_MACHINE, regEntry, 0, keyClass, 0, KEY_WRITE, 0, &phkResult, &dwDisposition);
    if (result != ERROR_SUCCESS)
    {
        result = ::RegCreateKeyExW(HKEY_CURRENT_USER, regEntry, 0, keyClass, 0, KEY_WRITE, 0, &phkResult, &dwDisposition);
        if (result != ERROR_SUCCESS)
        {
            unguard();
            return;
        }
    }

    //::RegSetValueExW(phkResult, keyStr, 0, REG_BINARY, (const BYTE *)buffer, 256u);
    ::RegSetValueExW(phkResult, keyStr, 0, REG_BINARY, (const BYTE*)buffer, sizeof(buffer));  // FIXED
    ::RegCloseKey(phkResult);

    unguard();
}
