#include "network/CLogCli.h"

#include "network/CLogSocket.h"
#include "network/packets/LogDPackets.h"
#include "utils/dbg/StackGuard.h"
#include "ui/CLog.h"
#include "config/Config.h"

CLogCli g_cliLog;  // 0x01BD4FA4

// 0x0040D1B0
CLogCli::CLogCli()
    : m_logSocket(NULL)
    , m_lock()
{
    guard(L"CLogCli::CLogCli()");

    ::InitializeCriticalSection(&m_lock);

    unguard();
}

// 0x00420A30
CLogCli::~CLogCli()
{
    ::DeleteCriticalSection(&m_lock);
}

// 0x0040C3F0
void CLogCli::Job()
{
    while (true)
    {
        g_cliLog.Init();
        ::SuspendThread(g_winlog.cliLogThread());
    }
}

// 0x0040D3C0
bool CLogCli::Init()
{
    guard(L"bool CLogCli::Init()");  // FIXED: added

    ::EnterCriticalSection(&m_lock);

    if (g_winlog.IsSendable())
    {
        ::LeaveCriticalSection(&m_lock);
        unguard();
        return true;
    }
    g_winlog.Add(LOG_INF, L"CLogCli::Init()  Connect()");
    if (Connect())
    {
        g_winlog.Sendable(true);
        ::LeaveCriticalSection(&m_lock);
        unguard();
        return true;
    }

    g_winlog.Sendable(false);

    ::LeaveCriticalSection(&m_lock);

    unguard();
    return false;
}

// 0x0040D1F0
bool CLogCli::Connect()
{
    guard(L"bool CLogCli::Connect()");

    SOCKET socket = ::socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
    sockaddr_in name;
    name.sin_family = AF_INET;
    name.sin_addr = Config::s_ipAddr;
    name.sin_port = htons(Config::s_port);

    if (::connect(socket, (const struct sockaddr*)&name, sizeof(sockaddr_in)) == SOCKET_ERROR)
    {
        ::closesocket(socket);
        g_winlog.Add(LOG_ERROR, L"Can't connect to log server");
        unguard();
        return false;
    }

    g_winlog.Add(LOG_INF, L"Connected to log server");

    m_logSocket = new CLogSocket(socket, CLogSocket::s_packetHandlers);

    m_logSocket->Initialize();
    m_logSocket->Send("cd", LogD_SetCheckStatus, 1);

#ifdef L2CACHED
    m_logSocket->Send("cdd", LogD_ServerStarted, Config::s_worldId, ServerType_CacheD);
#elif defined L2LOGD
    m_logSocket->Send("cdd", LogD_ServerStarted, Config::s_worldId, ServerType_LogD);
#endif  // #ifdef L2CACHED

    unguard();

    return true;
}

// 0x0040D350
void CLogCli::LogSendMsg(LogType logType, const wchar_t* msg)
{
    guard(L"void CLogCli::LogSendMsg(LogType nLogType, WCHAR* szSendMsg)");

#ifdef L2CACHED
    m_logSocket->Send("cddS", LogD_Send, ServerType_CacheD, logType, msg);
#elif defined L2LOGD
    m_logSocket->Send("cddS", LogD_Send, ServerType_LogD, logType, msg);
#endif  // #ifdef L2CACHED

    unguard();
}
