#include "network/CPacket.h"

#include "network/CPacketPool.h"
#include "network/CSlot.h"
#include "network/CIOBuffer.h"
#include "network/CIOSocket.h"
#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"
#include "ui/CLog.h"
#include "memory/Allocator.h"

// 0x004207F0
long CPacket::s_nAlloc = ::InterlockedIncrement(&MemoryObject::s_refCountSlots);  // 0x01BD3E0C

long CPacket::g_nPendingPacket;  // 0x01BD3BB0

// inlined
CPacket::CPacket()
    : CIOObject()
    , m_pSocket(NULL)
    , m_pBuf(NULL)
    , m_pFunc(NULL)
{
}

// inlined
CPacket::~CPacket()
{
}

// inlined in CLogSocket::OnRead
void* CPacket::operator new(size_t size)
{
    void* memory = g_allocator.allocate(size, 1, L"CPacket");

    ::InterlockedIncrement(&Allocator::g_allocatedObjects[CPacket::s_nAlloc]);

    Allocator::g_allocatedObjName[CPacket::s_nAlloc] = "CPacket";
    Allocator::g_allocatedMemory[CPacket::s_nAlloc] = sizeof(CPacket);

    return memory;
}

// 0x004092C0
void CPacket::operator delete(void* /*ptr*/)
{
    g_winlog.Add(LOG_ERROR, L"Never Call Oprator Delete!!! CPacket at file[%s], line[%d]", TEXT(__FILE__), __LINE__);
    CallStack::ReportCallStack();
}

// inlined
CPacket* CPacket::Alloc()
{
    return new CPacket();
}

// 0x00409B50
void CPacket::FreeAll()
{
    guard(L"void CPacket::FreeAll()");

    for (int i = 0; i < CPacketPool::s_size; ++i)
    {
        CSlot& slot = CPacketPool::s_slots[i];
        slot.m_lock.Enter();

        CPacket* pPacket;
        while ((pPacket = slot.m_data.packet) != NULL)
        {
            slot.m_data.socket = pPacket->m_pSocket;
            delete pPacket;  // TODO: allocator.deallocate to be used
        }

        slot.m_lock.Leave();
    }

    unguard();
}

// 0x00409A20
void CPacket::Release(const char* file, int line, ORTs /*nWorkType*/, bool /*flag*/)
{
    long refCountAfterRelease = ::InterlockedDecrement(&m_nRefCount);
    if (refCountAfterRelease > 0)
    {
        return;
    }

    if (refCountAfterRelease != 0)
    {
        if (refCountAfterRelease > -1)
        {
            return;
        }

        g_winlog.Add(LOG_ERROR, "Invalid Memory (%d) on CPacket(%d)", refCountAfterRelease, m_identity);
        if (file)
        {
            g_winlog.Add(LOG_ERROR, "           Caller : file[%s], line[%d]", file, line);
        }

        ReportOrt(0, 0);
        CallStack::ReportCallStack();
        return;
    }

    if (inTimer())
    {
        if (file)
        {
            g_winlog.Add(LOG_ERROR, "deleting in-timer-object![CPacket] at file[%s], line[%d]", file, line);
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CPacket::s_nAlloc]);
        }
        else
        {
            g_winlog.Add(LOG_ERROR, "deleting in-timer-object![CPacket]");
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CPacket::s_nAlloc]);
        }
        return;
    }

    this->~CPacket();

    g_allocator.deallocate(this);

    ::InterlockedDecrement(&Allocator::g_allocatedObjects[CPacket::s_nAlloc]);
}

// 0x0040A460
void CPacket::OnIOCallback(BOOL /*bSuccess*/, DWORD dwTransferred, LPOVERLAPPED /*lpOverlapped*/)
{
    guard(L"void CPacket::OnIOCallback(BOOL bSuccess, DWORD dwTransferred, LPOVERLAPPED lpOverlapped)");

    guard(L"cpacket::oniocallback_funccall");

    uint8_t* packet = m_pBuf->m_Buffer + dwTransferred;
    if ((*m_pFunc)(m_pSocket, packet + 1))
    {
        m_pSocket->CIOSocket::Close();
    }

    unguard();

    m_pSocket->Release(__FILE__, __LINE__, ORT_IO_RETAIN, false);
    m_pBuf->Release();

    ::InterlockedDecrement(&CPacket::g_nPendingPacket);
    Release(__FILE__, __LINE__, ORT_CREATE_DELETE, false);

    unguard();
}
