#include "threads/CIODeadlockChecker.h"

#include "utils/dbg/StackGuard.h"
#include "utils/dbg/CallStack.h"
#include "memory/Allocator.h"
#include "ui/CLog.h"

CIODeadlockChecker g_deadlock_checker;  // 0x01BD3D5C

// 0x00420760
long CIODeadlockChecker::s_nAlloc = ::InterlockedIncrement(&MemoryObject::s_refCountSlots);  // 0x01BD3D78

// 0x004092A0
CIODeadlockChecker::CIODeadlockChecker()
    : CIOObject()
    , m_nLastTick(0)
{
}

// 0x00409CD0
CIODeadlockChecker::~CIODeadlockChecker()
{
}

// inlined inside 0x00409CD0
void CIODeadlockChecker::operator delete(void* /*ptr*/)
{
    g_winlog.Add(LOG_ERROR, L"Never Call Oprator Delete!!! CIODeadlockChecker at file[%s], line[%d]", TEXT(__FILE__), __LINE__);
    CallStack::ReportCallStack();
}

// inlined
void CIODeadlockChecker::Init()
{
    TimerExpired(0);
}

long CIODeadlockChecker::lastTick() const
{
    return m_nLastTick;
}

// 0x0040BBB0
void CIODeadlockChecker::TimerExpired(int /*id*/)
{
    SetShortTimeStamp(&m_nLastTick, 0);
    AddTimer(CIO_DEADLOCK_CHECKER_INTERVAL, 0);
}

// 0x00409330
void CIODeadlockChecker::Release(const char* file, int line, ORTs /*nWorkType*/, bool /*flag*/)
{
    long refCountAfterRelease = ::InterlockedDecrement(&m_nRefCount);
    if (refCountAfterRelease > 0)
    {
        return;
    }

    if (refCountAfterRelease != 0)
    {
        if (refCountAfterRelease > -1)
        {
            return;
        }

        g_winlog.Add(LOG_ERROR, "Invalid Memory (%d) on CIODeadlockChecker(%d)", refCountAfterRelease, m_identity);
        if (file)
        {
            g_winlog.Add(LOG_ERROR, "           Caller : file[%s], line[%d]", file, line);
        }

        ReportOrt(0, 0);
        CallStack::ReportCallStack();
        return;
    }

    if (inTimer())
    {
        if (file)
        {
            g_winlog.Add(LOG_ERROR, "deleting in-timer-object![CIODeadlockChecker] at file[%s], line[%d]", file, line);
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CIODeadlockChecker::s_nAlloc]);
        }
        else
        {
            g_winlog.Add(LOG_ERROR, "deleting in-timer-object![CIODeadlockChecker]");
            ReportOrt(0, 0);
            ::InterlockedDecrement(&Allocator::g_allocatedObjects[CIODeadlockChecker::s_nAlloc]);
        }
        return;
    }

    this->~CIODeadlockChecker();

    g_allocator.deallocate(this);

    ::InterlockedDecrement(&Allocator::g_allocatedObjects[CIODeadlockChecker::s_nAlloc]);
}
