#pragma once

#include "logger/LogType.h"
#include "logger/ServerTypes.h"
#include "logger/FileBuffers.h"
#include "threads/ThreadIndices.h"

#include <windows.h>
#include <varargs.h>

class CLog
{
public:
    CLog(int size);  // 0x0040CF20
    ~CLog();         // 0x00420A20

    void Set(HWND logWnd);                                   // 0x0040BDE0
    void Redraw();                                           // 0x0040BF30
    void Enable(bool enable);                                // 0x0040C0A0
    void Sendable(bool sendable);                            // 0x0040C0E0
    bool IsSendable() const;                                 // inlined
    void FlushLogFile();                                     // 0x0040C110
    void Add(LogType severity, const char* format, ...);     // 0x0040D060
    void Add(LogType severity, const wchar_t* format, ...);  // 0x0040D0D0

    void ForceNewFile();  // inlined in binary
    HANDLE cliLogThread() const;

private:
    bool CanAppend(int year, int month, int day, int hour, bool half) const;   // 0x0040BE80
    void CreateNewLogFile(int year, int month, int day, int hour, bool half);  // 0x0040C180
    void LockBuffer(int bufferIndex);                                          // 0x0040BEF0
    void UnlockBuffer(int bufferIndex);                                        // 0x0040BF10
    void AddV(LogType severity, const char* format, va_list params);           // 0x0040C4C0
    void AddV(LogType severity, const wchar_t* format, va_list params);        // 0x0040CA10

private:
    struct LogEntry;
    struct FileName
    {
        wchar_t path[MAX_PATH];
    };

private:
    static const wchar_t* s_serverTypes[7];             // 0x0042D860
    static const wchar_t* s_logCategories[8];           // 0x0042D87C
    static int s_stackSizes[ThreadIndex_MaxCount + 1];  // 0x01BD3E18

    HWND m_window;
    bool m_enabled;
    bool m_sendable;
    // bool m_enabled3; unused
    const int m_bufferLength;
    int m_currentLine;
    LogEntry* m_outputBuffer;
    CRITICAL_SECTION m_criticalSection;
    RECT m_loginWndRect;
    SIZE m_textSize;
    HGDIOBJ m_font;
    HBRUSH m_brush;
    unsigned int m_threadAddr;
    HANDLE m_cliLogThread;
    int m_year;
    int m_month;
    int m_day;
    int m_hour;
    bool m_half;
    int m_worldID;
    ServerTypes m_logComponent;
    HANDLE m_fileBuffers[FileBuffers_BuffersCount];
    CRITICAL_SECTION m_bufferLocks[FileBuffers_BuffersCount];
    FileName m_fileNames[FileBuffers_BuffersCount];
    bool m_createNewFile;
};

extern CLog g_winlog;  // 0x01BD3E50
